#include "compress.h"

//*************************************************************
// Must be in lexicographic order!

char *code_tab2[]={		"0",
						"1"
};

char *code_tab3[]={		"0",
						"10",
						"11"
};
char *code_tab7[]={		"0",		// 0
						"100",		// 1
						"101",		// 2
						"1100",		// 3
						"1101",		// 4
						"1110",		// 5
						"1111"		// 6
};

char *code_tab17[]={	"0",		// 0
						"100",		// 1
						"101",		// 2
						"11000",	// 3
						"11001",	// 4
						"110100",	// 5
						"110101",	// 6
						"110110",	// 7
						"110111",	// 8
						"1110000",	// 9
						"1110001",	// 10
						"1110010",	// 11
						"1110011",	// 12
						"1110100",	// 13
						"1110101",	// 14
						"1110110",	// 15
						"1110111",	// 16
						"11110000",	// 17
						"11110001",	// 18
						"11110010",	// 19
						"11110011", // 20
						"11110100", // 21
						"11110101", // 22
						"11110110", // 23
						"11110111", // 24
						"11111000", // 25
						"11111001", // 26
						"11111010", // 27
						"11111011",	// 28
						"11111100", // 29
						"11111101", // 30
						"11111110", // 31
						"11111111"  // 32
};
char **code_table;


//*************************************************************

Compressor::CompDictionary::CompDictionary()
{
	root=NULL;
	real_dict_size=0;
	total=0;
}

Compressor::CompDictionary::~CompDictionary()
{
	delete root;
}

int Compressor::CompDictionary::Add(int x)
{
	if(root==NULL) 
	{
		root=new CompTree(x);
		tree_size=1;
	}
	else tree_size+=root->Add(x);
	return (tree_size>DICT_SIZE);
}

void Compressor::CompDictionary::Add(CompTree *p)
{
	if(p==NULL) return;
	AddToDictionary(p->value,p->number);
	Add(p->left);
	Add(p->right);
}

void Compressor::CompDictionary::AddToDictionary(int val,int freq)
{
	if(val!=OTHER_VALS) total+=freq;
	int i=0;
	while(i<real_dict_size && freq<=nums[i]) i++;
	if(i==real_dict_size)
	{
		if(real_dict_size<DICT_SIZE)
		{
			real_dict_size++;
			nums[i]=freq;
			vals[i]=val;
		}
		else if(val==OTHER_VALS)
		{
			nums[DICT_SIZE-1]=freq;
			vals[DICT_SIZE-1]=val;
		}
	}
	else
	{
		if(real_dict_size<DICT_SIZE) real_dict_size++;
		for(int j=real_dict_size-2;j>=i;j--)
		{
			nums[j+1]=nums[j];
			vals[j+1]=vals[j];
		}
		nums[i]=freq;
		vals[i]=val;
	}
}

double Compressor::CompDictionary::CreateDictionary()
{
	Add(root);
	int sum=0;
	if(real_dict_size==DICT_SIZE)
	{
		for(int i=0;i<DICT_SIZE;i++) sum+=nums[i];
		if(total>sum) AddToDictionary(OTHER_VALS,total-sum);
	}
	else sum=total;
	other_vals_index=OTHER_VALS;
	for(int i=0;i<real_dict_size;i++) if(vals[i]==OTHER_VALS) other_vals_index=i;
	return (total-sum)/(double)total;
}

int Compressor::CompDictionary::Encode(int n)
{
	for(int i=0;i<real_dict_size;i++) if(vals[i]==n) return i;
	return other_vals_index;
}


//*************************************************************

Compressor::CompTree::CompTree(int x)
{
	value=x;
	number=1;
	left=right=NULL;
}

Compressor::CompTree::~CompTree()
{
	delete right;
	delete left;
}

int Compressor::CompTree::Add(int x)
{
	if(x==value) 
	{
		number++;
		return 0;
	}
	if(x<value)
	{
		if(left==NULL) left=new CompTree(x);
		else return left->Add(x);
		return 1;
	}
	if(right==NULL) right=new CompTree(x);
	else return right->Add(x);
	return 1;
}

//***************************************************************

Compressor::Compressor(int fhandle,int mode_)
{
	fhnd=fhandle;
	buffer=NULL;
	b_len=0;
	b_cur_pos=0;

	buf=0;
	cur_bit=0;
	mode=mode_;	// 0 - write, 1 - read
	code_table=code_tab17;
	repetition=0;
	prev_val=OTHER_VALS;
	p2=3;
	p5=3;
	min=2000000000;
	max=-min;
	rep_flag=0;
	total=0;
	dic=NULL;
}

Compressor::Compressor(int *b,int len_,int mode_)
{
	fhnd=-1;
	buffer=b;
	b_len=len_;
	b_cur_pos=0;

	buf=0;
	cur_bit=0;
	mode=mode_;	// 0 - write, 1 - read
	code_table=code_tab17;
	repetition=0;
	prev_val=OTHER_VALS;
	p2=3;
	p5=3;
	min=2000000000;
	max=-min;
	rep_flag=0;
	total=0;
	dic=NULL;
}

Compressor::~Compressor()
{
	if(mode==0) Flush();
	delete dic;
}

int Compressor::Flush()
{
	if(mode==0 && cur_bit>0)
	{
		if(fhnd>-1) write(fhnd,&buf,sizeof(unsigned int));
		else
		{
			if(b_cur_pos>b_len-1) throw CompressionFailed();
			buffer[b_cur_pos++]=buf;
		}
	}
	buf=0;
	cur_bit=0;
	return b_cur_pos;
}

void Compressor::Save(char *str)
{
	assert(mode==0);
	while(*str!='\0')
	{
		if(*str=='1') buf+=1 << cur_bit;
		cur_bit++;
		str++;
		if(cur_bit==32)
		{
			if(mode!=0) throw CompressionFailed();
			Flush();
		}
	}
}

void Compressor::Save(int n)
{
	assert(mode==0);
	if(n) buf+=1 << cur_bit;
	cur_bit++;
	if(cur_bit==32) 
	{
		if(mode!=0) throw CompressionFailed();
		Flush();
	}
}

int Compressor::Load()
{
	assert(mode==1);
	if(mode!=1) throw DecompressionFailed();
	if(cur_bit==0)
	{
		if(fhnd>-1) read(fhnd,&buf,sizeof(unsigned int));
		else
		{
			buf=buffer[b_cur_pos++];
		}
	}
	else buf=buf/2;
	cur_bit++;
	if(cur_bit==32) cur_bit=0;
	return buf%2;
}

int Compressor::Eof()
{
	if(fhnd==-1) return 0;
	if(eof(fhnd)!=0 && cur_bit==0) return 1;
	return 0;
}

void Compressor::SaveNumber(int nn)
{
	unsigned int n=(unsigned int)nn;
	if(n<(1<<SHORT_SIZE))
	{
		Save(0);		// short number indicator
		for(int i=0;i<SHORT_SIZE;i++)
		{
			Save(n%2);
			n/=2;
		}
	}
	else
	{
		Save(1);		// long number indicator
		for(int i=0;i<32;i++)
		{
			Save(n%2);
			n/=2;
		}
	}
}

int Compressor::LoadNumber()
{
	unsigned int n=0;
	int l=Load();
	if(l==0)		
	{				// short number
		for(int i=0;i<SHORT_SIZE;i++)
		{
			if(Load()==1) n+=(1<<i);
		}
	}
	else
	{
		for(int i=0;i<32;i++)
		{
			if(Load()==1) n+=(1<<i);
		}
	}
	return (int)n;
}

void Compressor::SaveQuant(int n)	// n from 2 to 2+2^LONG_QUANT+2^SHORT_QUANT
{
	if(n<2+(1<<SHORT_QUANT))
	{
		n-=2;
		Save(0);		// short number indicator
		for(int i=0;i<SHORT_QUANT;i++)
		{
			Save(n%2);
			n/=2;
		}
	}
	else if(n<2+(1<<SHORT_QUANT)+(1<<LONG_QUANT))
	{
		n-=2+(1<<SHORT_QUANT);
		Save(1);		// long number indicator
		for(int i=0;i<LONG_QUANT;i++)
		{
			Save(n%2);
			n/=2;
		}
	}
	else throw CompressionFailed();
}

int Compressor::LoadQuant()
{
	int n=0;
	int l=Load();
	if(l==0)		
	{				// short number
		for(int i=0;i<SHORT_QUANT;i++)
		{
			if(Load()==1) n+=(1<<i);
		}
		n+=2;
	}
	else
	{
		for(int i=0;i<LONG_QUANT;i++)
		{
			if(Load()==1) n+=(1<<i);
		}
		n+=2+(1<<SHORT_QUANT);
	}
	return n;
}

void Compressor::SaveDictVal(int n)
{
	int code;
	if(dic)
	{
		code=dic->Encode(n);
		SaveDict(code);
	}
	if(dic==NULL || dic->vals[code]==OTHER_VALS) 
		if(min<2000000000) SaveVal((n-min)/divide,diff_bit);
		else SaveNumber(n);
}

int Compressor::LoadDictVal()
{
	int code,val;
	if(dic)
	{
		code=LoadDict();
		val=dic->vals[code];
	}
	if(dic==NULL || dic->vals[code]==OTHER_VALS) 
		if(min<2000000000) val=min+divide*LoadVal(diff_bit);
		else val=LoadNumber();
	return val;
}

void Compressor::SaveDict(int n)	// n up to DICT_SIZE
{
	assert(n<DICT_SIZE);
	if(n>=DICT_SIZE || n<0) throw CompressionFailed();
	Save(code_table[n]);
}

int Compressor::LoadDict()
{
	int n=0;
	int k=0;
	int l;
	int ld=1;
	while(n<DICT_SIZE-1 && code_table[n][k]!='\0')
	{
		if(ld) l=Load();
		ld=0;
		if((l==1 && code_table[n][k]=='1')||(l==0 && code_table[n][k]=='0')) 
		{
			k++;
			ld=1;
		}
		else n++;
	}
	return n;
}

void Compressor::SaveVal(int n,int k)	// code n on k bits
{
	for(int i=0;i<k;i++)
	{
		Save(n%2);
		n/=2;
	}
	assert(n==0);
	if(n>0) throw CompressionFailed();
}

int Compressor::LoadVal(int k)
{
	int n=0;
	for(int i=0;i<k;i++)
	{
		if(Load()==1) n+=(1<<i);
	}
	return n;
}

void Compressor::SaveStr(char *buf)
{
	int seven=1,len=0;
	while(buf[len]!='\0') 
	{
		if(buf[len]>127 || buf[len]<0) seven=0;
		len++;
	}
	Save(seven);
	seven=(seven?7:8);
	for(int i=0;i<=len;i++) SaveVal((unsigned char)(buf[i]),seven);  // last char=0
}

int Compressor::LoadStr(char *buf,int max_len)
{
	int seven,len=0,last;
	seven=Load();
	seven=(seven?7:8);
	do
	{
		last=LoadVal(seven);
		if(len<max_len) buf[len]=char(last);
		if(len==max_len) buf[len]='\0';
		len++;
	}
	while(last>0);
	return 0;
}

void Compressor::SaveFloat(double v)
{
	if(v>=0 && v<(1<<SHORT_FLOAT) && int(v)==v)
	{
		Save(0);
		SaveVal(int(v),SHORT_FLOAT);
	}
	else
	{
		Save(1);
		int *p=(int*)(&v);
		int s=sizeof(double)/4;
		for(int i=0;i<s;i++) SaveVal(p[i],32);
	}
}

double Compressor::LoadFloat()		
{
	double v;
	if(Load()==0)
	{
		v=LoadVal(SHORT_FLOAT);
	}
	else
	{
		int *p=(int*)(&v);
		int s=sizeof(double)/4;
		for(int i=0;i<s;i++) p[i]=LoadVal(32);
	}
	return v;
}

int power_5[]={5,25,625};

void Compressor::InitSaveStream(int no_rep)
{
	divide=1;
	if(p5>0) divide*=power_5[p5-1];
	if(p2>0) divide*=(1<<p2);
	unsigned int diff=(max-min)/divide+1;
	for(int i=0;diff>0;i++) diff/=2;
	diff_bit=i;
	if(repetition>total/4 && no_rep==0) rep_flag=1;
	else rep_flag=0;
	repetition=0;

	double other_percent=0,coefficient=1;
	other_percent=dic->CreateDictionary();
	if(other_percent>0) 
		if(diff_bit>5) coefficient=(diff_bit-5)/double(diff_bit-5+1);
			else coefficient=0;

	if(other_percent>coefficient)
	{
		delete dic;
		dic=NULL;
	}

	int rsize;
	if(dic) rsize=dic->GetRealSize();
	else rsize=0;

	SaveQuant(rsize+2);	// dict. size
	for(i=0;i<rsize;i++) SaveNumber(dic->vals[i]); // dictionary
	if(rep_flag) Save(1);	// repetition flag
	else Save(0);
	SaveNumber(min);		// shift of OTHER_VALS
	SaveQuant(diff_bit);	// bit size of OTHER_VALS
	SaveVal(p2,2);			// power of 2 in scale coefficient
	SaveVal(p5,2);			// power of 5 in scale coefficient

	repetition=0;
	prev_val=OTHER_VALS;
	if(rsize==2) code_table=code_tab2;
	else if(rsize==3) code_table=code_tab3;
	else if(rsize<8) code_table=code_tab7;
	else code_table=code_tab17;
}

void Compressor::SaveStream(int n)
{
	if(rep_flag)
	{
		if(prev_val==OTHER_VALS)		// beginning of the stream
		{
			prev_val=n;
			repetition=1;
		}
		else if(prev_val==n && repetition<(1<<LONG_QUANT)+1+(1<<SHORT_QUANT)) 
			repetition++;
		else
		{
			if(repetition>1) 
			{
				Save(1);	// repetition flag
				SaveQuant(repetition);
			}
			else Save(0);
			SaveDictVal(prev_val);
			repetition=1;
			prev_val=n;
		}
	}
	else 
	{
		SaveDictVal(n);
	}
}

void Compressor::FlushSaveStream()
{
	if(rep_flag) SaveStream(OTHER_VALS);
	repetition=0;
	prev_val=OTHER_VALS;
	p2=3;
	p5=3;
	min=2000000000;
	max=-min;
	rep_flag=0;
	total=0;
	delete dic;
	dic=NULL;
}

void Compressor::InitLoadStream()
{
	delete dic;
	int dic_real_size=LoadQuant()-2;
	if(dic_real_size>0) 
	{
		dic=new CompDictionary();
		dic->SetRealSize(dic_real_size);
	}
	else dic=NULL;
	for(int i=0;i<dic_real_size;i++) dic->vals[i]=LoadNumber();
	rep_flag=Load();
	repetition=0;
	prev_val=OTHER_VALS;
	min=LoadNumber();
	diff_bit=LoadQuant();
	p2=LoadVal(2);
	p5=LoadVal(2);
    divide=1;
	if(p5>0) divide*=power_5[p5-1];
	if(p2>0) divide*=(1<<p2);
	if(dic_real_size==2) code_table=code_tab2;
	else if(dic_real_size==3) code_table=code_tab3;
	else if(dic_real_size<8) code_table=code_tab7;
	else code_table=code_tab17;
}

int Compressor::LoadStream()
{
	if(rep_flag)
	{
		if(repetition>0) 
		{
			repetition--;
			return prev_val;
		}
		int r=Load();
		if(r) repetition=LoadQuant()-1;
		else repetition=0;
		prev_val=LoadDictVal();
		return prev_val;
	}
	prev_val=LoadDictVal();
	return prev_val;
}

void Compressor::PrepareStream(int val)
{
	total++;
	if(dic==NULL) dic=new CompDictionary();	
	dic->Add(val);
	if(val==prev_val) repetition++;
	prev_val=val;
	if(val<min) min=val;
	if(val>max) max=val;
	for(int i=0;i<p5;i++) if(val%power_5[i]!=0) p5=i;
	for(i=0;i<p2;i++) if(val%(2<<i)!=0) p2=i;
}

