// ************** TDTable - ver. 2.21 *******************
#include "tdtable.h"

TDTable::TDTable()
{
  tab=NULL;
  l_obj=l_attr=0;
  attr_dec=-1;
}

TDTable::TDTable(int o,int a)
{
  l_obj=o;
  l_attr=a;
  tab=new int [l_obj*l_attr];
  attr_dec=-1;
}

TDTable::~TDTable()
{
  delete [] tab;
}

int TDTable::GetTable(int o,int a) 
{
	  assert(o>=0 && o<l_obj && a>=0 && a<l_attr);
	  return tab[o*l_attr+a]; 
}

void TDTable::SetTable(int o,int a,int val) 
{
	  assert(o>=0 && o<l_obj && a>=0 && a<l_attr);
	  tab[o*l_attr+a]=val; 
}

int	TDTable::GetDecAttr()
{ 
	assert(attr_dec>=0);
	return attr_dec; 
}

void TDTable::SetDecAttr(int a)
{ 
	attr_dec=a; 
}


int *TDTable::GetTable(int o)     
{ 
	  
	  return tab+o*l_attr; 
}

int TDTable::LoadInfo(char *name)	// return: 0 - ok, 1 - error
{
	tab=NULL;
	int result=1,compress=1;

	int fhandle=open(name,O_BINARY | O_RDONLY);
	if(fhandle==-1) result=0;
	char buf[5];
	if(result)
	{
		read(fhandle,buf,4);
		if(buf[0]!='R' || buf[1]!='S' || buf[2]!='c') compress=0;
		if(buf[3]<'0' || buf[3]>'0') compress=0;
		if(result && compress)
		{
			Compressor c(fhandle,1);
			l_obj=c.LoadNumber();
			l_attr=c.LoadNumber();
			if((l_obj<=0)||(l_attr<=0)) result=0;
		}
		close(fhandle);
	}
	if(result && compress==0)
	{
		ifstream f;
		f.open(name,ios::in);
		if(!f) result=0;
		f >> l_obj;
		f >> l_attr;
		if((l_obj<=0)||(l_attr<=0)) result=0;
		f.close();
	}
	return 1-result;
}


int TDTable::LoadTable(char *name,int n_ob,int *o)	// return: 0 - ok, 1 - error
{
	int fhandle=open(name,O_BINARY | O_RDONLY);
	if(fhandle==-1) return 1;

	char buf[5];
	read(fhandle,buf,4);
	int result=0;
	if(buf[0]!='R' || buf[1]!='S' || buf[2]!='c') result=1;
	if(buf[3]<'0' || buf[3]>'0') result=1;
	if(result==0)
	{
		Compressor c(fhandle,1);
		result=Decompress(c,n_ob,o);
	}

	close(fhandle);
	if(result==0) return 0;

	ifstream f;
	f.open(name,ios::in);
	assert(f);
	f >> l_obj;
	f >> l_attr;
	if((l_obj<=0)||(l_attr<=0)) return 1;
	delete [] tab;
	char buf1[30];
	try
	{
	if(n_ob==0)
	{
		tab=new int [l_obj*l_attr];
		for(int i=0;i<l_obj;i++)
		{
			for(int j=0;j<l_attr;j++) 
			{
				f >> buf1;
				if(buf1[0]=='?') tab[i*l_attr+j]=NULL_VALUE;
				else tab[i*l_attr+j]=int(atof(buf1));
			}
//			::Percent((i+1)/double(l_obj));
		}
	}
	else
	{
		tab=new int [n_ob*l_attr];
		int k=0,buf;
		for(int i=0;i<l_obj && k<n_ob;i++)
		{
//			::Percent((i+1)/double(l_obj));
			if(i==o[k])
			{
				for(int j=0;j<l_attr;j++) 
				{
					f >> buf; 
					tab[k*l_attr+j]=buf;
				}
				k++;
			}
			else for(int j=0;j<l_attr;j++) f >> buf; 
		}
		l_obj=n_ob;
	}
	}
	catch(...)
	{
		f.close();
		attr_dec=-1;
		l_obj=0;
		l_attr=0;
		delete [] tab;
		tab=NULL;
		throw;
	}
	f.close();
	attr_dec=l_attr-1;
	return 0;
}

int TDTable::SaveTable(char *name,int compress,int n_ob,int *o)	
// return: 0 - ok, 1 - error
{
	if(n_ob<0 || n_ob>=NoObj()) n_ob=0;
	if(compress)
	{
		int fhandle=open(name,O_BINARY | O_CREAT | O_EXCL | O_WRONLY, S_IREAD | S_IWRITE);
		if(fhandle==-1)
		{	// truncate
			close(fhandle);
			fhandle=open(name,O_BINARY | O_TRUNC | O_WRONLY, S_IREAD | S_IWRITE);
		}
		if(fhandle==-1) return 1;
		write(fhandle,"RSc0",4);
		Compressor c(fhandle,0);
		Compress(c,n_ob,o);
		close(fhandle);
		return 0;
	}

	ofstream f;
	f.open(name,ios::out);
	assert(f);
	if(n_ob==0)
	{
		f << l_obj << " ";
		f << l_attr << endl;
		if((l_obj<=0)||(l_attr<=0)) return 1;
		for(int i=0;i<l_obj;i++)
		{
		  for(int j=0;j<l_attr;j++) f << tab[i*l_attr+j] << " ";
		  f << endl;
		}
	}
	else
	{
		int k=0;
		f << n_ob << " ";
		f << l_attr << endl;
		if((l_obj<=0)||(l_attr<=0)) return 1;
		for(int i=0;i<l_obj && k<n_ob;i++) if(i==o[k])
		{
		  for(int j=0;j<l_attr;j++) f << tab[i*l_attr+j] << " ";
		  f << endl;
		  k++;
		}
	}
	f.close();
	if(f) return 0;
	return 1;
}

TDTable::TDTable(TDTable *table)
{
	l_attr=table->NoAttr();
	l_obj=table->NoObj();
	attr_dec=table->attr_dec;
	tab=new int [l_obj*l_attr];
	for(int i=0;i<l_obj*l_attr;i++)
		tab[i]=table->tab[i];
}

TDTable::TDTable(TDTable *table,int n,int *o)
{
	l_attr=table->NoAttr();
	l_obj=n;
	attr_dec=table->attr_dec;
	tab=new int [l_obj*l_attr];
	for(int i=0;i<l_obj;i++)
	{
		for(int j=0;j<l_attr;j++) 
			tab[i*l_attr+j]=table->tab[o[i]*l_attr+j];
	}
}

int TDTable::Compress(Compressor &c,int n_ob,int *o)
{
	if(n_ob==0)
	{
		c.SaveNumber(NoObj());
		c.SaveNumber(NoAttr());
		for(int a=0;a<NoAttr();a++)
		{
//			cout << "\b\b\b\b\b\b" << ((a+1)*100)/l_attr << "%" << flush;
			for(int obj=0;obj<NoObj();obj++) c.PrepareStream(GetTable(obj,a));
			c.InitSaveStream();
			for(obj=0;obj<NoObj();obj++) 
			{
				c.SaveStream(GetTable(obj,a));
			}
			c.FlushSaveStream();
		}
	}
	else
	{
		c.SaveNumber(n_ob);
		c.SaveNumber(NoAttr());
		for(int a=0;a<NoAttr();a++)
		{
			int k=0;
//			cout << "\b\b\b\b\b\b" << ((a+1)*100)/l_attr << "%" << flush;
			for(int obj=0;obj<NoObj() && k<n_ob;obj++) 
			{
				if(o[k]==obj)
				{
					c.PrepareStream(GetTable(obj,a));
					k++;
				}
			}
			k=0;
			c.InitSaveStream();
			for(obj=0;obj<NoObj() && k<n_ob;obj++) 
			{
				if(o[k]==obj)
				{
					c.SaveStream(GetTable(obj,a));
					k++;
				}
			}
			c.FlushSaveStream();
		}
	}
	c.Flush();
//	cout << endl;
	return 1;
}

int TDTable::Decompress(Compressor &c,int n_ob,int *o)
{
	l_obj=c.LoadNumber();
	l_attr=c.LoadNumber();
	if((l_obj<=0)||(l_attr<=0)) return 1;
	delete [] tab;
	attr_dec=l_attr-1;
	if(n_ob==0)
	{
		tab=new int [l_obj*l_attr];
		for(int a=0;a<NoAttr();a++)
		{
//			::Percent((a+1)/double(l_attr));
//			cout << "\b\b\b\b\b\b" << ((a+1)*100)/l_attr << "%" << flush;
			c.InitLoadStream();
			for(int obj=0;obj<NoObj();obj++) SetTable(obj,a,c.LoadStream());
		}
	}
	else
	{
		tab=new int [n_ob*l_attr];
		for(int a=0;a<NoAttr();a++)
		{
//			::Percent((a+1)/double(l_attr));
//			cout << "\b\b\b\b\b\b" << ((a+1)*100)/l_attr << "%" << flush;
			int k=0;
			c.InitLoadStream();
			for(int obj=0;obj<l_obj;obj++)
				if(k<n_ob && obj==o[k]) SetTable(k++,a,c.LoadStream());
				else c.LoadStream();
		}
		l_obj=n_ob;
	}
//	cout << endl;
	return 0;
}

int TDTable::MemoryCompress(int *b,int len,int n,int *o)
{
	Compressor c(b,len,0);
	Compress(c,n,o);
	return c.Flush();
}

int TDTable::MemoryDecompress(int *b)
{
	Compressor c(b,0,1);
	return Decompress(c);
}


//*****************************************************************
//****** Sample generator files

SampleGenerator::SampleGenerator(int all,int part)
{
	o=new int [all];
	if(all<part) part=all;
	no_samples=all/part;
	n_all=all;
	n_part=part;
	int i;
	for(i=0;i<all;i++) o[i]=i;
	for(i=all-1;i>0;i--)
	{
		int r=(rand()*RAND_MAX+rand())%(i+1);
		int pom=o[i];
		o[i]=o[r];
		o[r]=pom;
	}
	for(i=0;i<no_samples;i++)
		QuickSortTable(i*part,(i+1)*part-1);
	o_pom=NULL;
}

SampleGenerator::~SampleGenerator()
{
	delete [] o;
	delete [] o_pom;
}

int *SampleGenerator::CVtest(int j)
{
	if(j>=no_samples) return NULL;
	return o+n_part*j;
}

int *SampleGenerator::CVtraining(int j)
{
	if(j>=no_samples) return NULL;
	delete [] o_pom;
	o_pom=new int [n_part*(no_samples-1)];
	int *cur=new int [no_samples-1];
	int ii=0;
	for(int i=0;i<no_samples;i++)
		if(j!=i)
			cur[ii++]=n_part*i;
	int dest=0;
	do				// merge sort
	{
		int min=n_all,min_pos=-1;
		ii=0;
		for(i=0;i<no_samples;i++) if(i!=j)
		{
			if(cur[ii]<n_part*(i+1) && o[cur[ii]]<min)
			{
				min=o[cur[ii]];
				min_pos=ii;
			}
			ii++;
		}
		o_pom[dest++]=min;
		cur[min_pos]++;
	}
	while(dest<n_part*(no_samples-1));
	delete [] cur;
	return o_pom;
}

void SampleGenerator::QuickSortTable(int l,int p)
// Quick Sort - see Wirth "Algorithms+data struct=prog", page 90.
{
  int srodek=o[(l+p)/2];
  int w,i=l,j=p;
  do
  {
	 while(o[i]<srodek) i++;
	 while(srodek<o[j]) j--;
	 if(i<=j)
	 {
		w=o[i];
		o[i]=o[j];
		o[j]=w;
		i++;
		j--;
	 };
  } while(i<=j);
  if(l<j) QuickSortTable(l,j);
  if(i<p) QuickSortTable(i,p);
}