//**************************** Fast TDTable object ************************
//************* (c) 1996 J. Wroblewski
//*** version 2.21 (compatible with RSES, compression, NULL values)

#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <stdlib.h>
#include <sys\stat.h>

#ifndef _TDTABLE_H_
#define _TDTABLE_H_ 1

#include "compress.h"

class SQLBase;
class SQLTable;
class GenPattern;

class Table
{
public:
	virtual int NoAttr()=0;
	virtual int NoObj()=0;
	virtual int GetTable(int,int)=0;
	virtual int GetDecAttr()=0;
};

class TDTable:public Table
{
public:
  TDTable();
  TDTable(int,int);
  TDTable(TDTable *);
  TDTable(TDTable *,int n,int *o);	// as LoadTable
  ~TDTable();

  int LoadTable(char *,int n=0,int *o=NULL);
		// if n>0:
		// loads only a part of table: n objects, numbers are taken
		// from o[] table. WARNING: o[] must be sorted ascending.
  int SaveTable(char *,int compress=0,int n=0,int *o=NULL);
		// if n>0 - see above.
  int LoadInfo(char *);		// loads no. of objects/attributes only

  int Compress(Compressor &,int n=0,int *o=NULL);
  int Decompress(Compressor &,int n=0,int *o=NULL);

  int MemoryCompress(int*,int,int n=0,int *o=NULL);	
		// buffer,size;  returns number of ints used
		// n,o - see "SaveTable"
  int MemoryDecompress(int*);	// buffer

  int GetTable(int o,int a);
  void SetTable(int o,int a,int val);
  int *GetTable(int o);     
  int NoObj()  { return l_obj; };
  int NoAttr() { return l_attr; };

  int GetDecAttr();
  void SetDecAttr(int a);

protected:
  int l_obj,l_attr;
  int *tab;
  int attr_dec;
};

//*********************************************************************************
//
//  Generates a table of "part" random numbers from 0 to "all",
//  sorted ascending.
//
//	Example:
//		TDTable t;
//		t.LoadTable("...");		// 1000 objects 
//		// Sampling: 50 from 1000:
//		SampleGenerator s(1000,50);	
//		t.Save("...",0,50,s.GetDescription());
//
//	Usage in CV-n mode:
//
//		TDTable t;
//		t.LoadTable("...");		// 1000 objects 
//		// Sampling: CV-5
//		SampleGenerator s(1000,200);	
//		t.Save("...",0,200,s.CVtest(0));
//		t.Save("...",0,800,s.CVtraining(0));
//		t.Save("...",0,200,s.CVtest(1));
//		t.Save("...",0,800,s.CVtraining(1));	// etc...

class SampleGenerator
{
public:
	SampleGenerator(int all,int part);
	~SampleGenerator();
	int *GetDescription()
	{ return o; }

	int *CVtest(int j);
	int *CVtraining(int j);
private:
	void QuickSortTable(int,int);
	int *o,*o_pom;
	int no_samples,n_all,n_part;
};


#endif
