
#include "subset.h"
//**************************************************************

SubsetTree::SubsetTree(Subset &s)
{
  x=s;   				// set root
  left=right=NULL;
};

SubsetTree::~SubsetTree()
{
  delete left;
  delete right;
  left=right=NULL;
};

int SubsetTree::AddIfNotExists(Subset &s)
{
  if(s==x) return 1;     // found?
  if(s<x)
  {
	 if(left==NULL)      // not found: insert
	 {
		left=new SubsetTree(s);
		if(left==NULL) throw MemError();
		return 0;
	 }
	 return left->AddIfNotExists(s);   // or continue searching
  }  // else
  if(right==NULL)      // not found: insert
  {
	 right=new SubsetTree(s);
	 if(right==NULL) throw MemError();
	 return 0;
  }
  return right->AddIfNotExists(s);   // or continue searching
}

int SubsetTree::Exists(Subset &s)
{
  if(s==x) return 1;			// as above, but does not insert
  if(s<x)
  {
	 if(left==NULL) return 0;
	 return left->Exists(s);
  } // else
  if(right==NULL) return 0;
  return right->Exists(s);
};

//**************************************************************

unsigned int Subset::mask[] = { 0x00000001,0x00000002,0x00000004,0x00000008,0x00000010,0x00000020,0x00000040,0x00000080,
		 0x00000100,0x00000200,0x00000400,0x00000800,0x00001000,0x00002000,0x00004000,0x00008000,
		 0x00010000,0x00020000,0x00040000,0x00080000,0x00100000,0x00200000,0x00400000,0x00800000,
		 0x01000000,0x02000000,0x04000000,0x08000000,0x10000000,0x20000000,0x40000000,0x80000000 };

Subset::Subset(int k)
{
  if(k>0)
  {
	 nr=(k-1)/32+1;		// calculate a number of 32-bit integers
	 a=new unsigned int [nr];
	 if(a==NULL) throw MemError();
	 for(int i=0;i<nr;i++) a[i]=0;
  }
  else
  { nr=0;
	 a=NULL;
	 cerr << "Subset 0 or less" << endl;
  };
  n=k;
};

Subset::Subset(Subset &sec)
{
  nr=sec.nr;		// copying constructor
  n=sec.n;
  a=new unsigned int [nr];
  if(a==NULL) throw MemError();
  for(int i=0;i<nr;i++) a[i]=sec.a[i];  // copy 32-bit blocks
};

void Subset::Zero()
{
  for(int i=0;i<nr;i++) a[i]=0;   // all bits in each block
};

void Subset::Full()
{
  for(int i=0;i<n;i++) Set(i);
};

Subset::~Subset()
{
  delete [] a;
  a=NULL;
};

int Subset::Ones()
{
  int s=0;
  for(int i=0;i<n;i++) s+=Val(i);
  return s;
};

int Subset::operator==(Subset& sec)
{
  if(n!=sec.n) cerr << "Subset: Operator== error" << endl;
  for(int i=0;i<nr;i++) if(a[i]!=sec.a[i]) return 0;
											// compare whole blocks
  return 1;
};

int Subset::operator<(Subset& sec)
{
  if(n!=sec.n) cerr << "Subset: Operator< error" << endl;
  for(int i=0;i<nr;i++)
  {
	 if(a[i]<sec.a[i]) return 1;	// compare whole blocks
	 if(a[i]>sec.a[i]) return 0;
  };
  return 0;
};

int Subset::operator<=(Subset& sec)
{
  if(n!=sec.n) cerr << "Subset: Operator<= error" << endl;
  for(int i=0;i<nr;i++) if((a[i]&(~sec.a[i]))!=0) return 0;
							// binary operation on whole blocks
  return 1;
};

Subset& Subset::operator=(Subset& sec)
{
  if(this!=&sec)
  {
	 delete [] a;
	 nr=sec.nr;
	 n=sec.n;
	 a=new unsigned int [nr];
	 if(a==NULL) throw MemError();
	 for(int i=0;i<nr;i++) a[i]=sec.a[i];
  };
  return *this;
};

ostream &operator<<(ostream &out,Subset &s)
{
  for(int i=0;i<s.n;i++) out << s[i];
  return out;
};

void Subset::Set(int *table)
{
  int i=0;				// table - a list of bit numbers to set,
  while((table[i]>-1)&&(table[i]<n))		// finished by -1
  {
	 Set(table[i]);
	 i++;
  }
};


// **************************************************************

