/*********************************************************

Name: subset.h
Author: JW
Revised by:
Version: 1.0.001
Date of mod.: 4.04.1998
Status: under development

Description:

Comments: requires 32-bit int.

*********************************************************/

//*************************************************************
// Contens:
//		class Subset:	a subset (characterist. function) of desired size
//          Subset(), copying constr., operator=, operator==,
//				operator< - lexicographical order,
//				operator<= - implication (set inclusion)
//				setting bits: Set(int), Reset(int), Zero(), Full().
//				getting bits:   operator[], Val(int)
//
//		class SubsetTree: a binary tree of Subsets (lexicographic order)
//				SubsetTree(Subset)  - root,
//				AddIfNotExists(Subset),
//				Exists(Subset)			- both returns 1 iff Subset was found on tree
//***************************************************************

#ifndef _SUBSET_H_
#define _SUBSET_H_

#include <iostream.h>
#include "rs_ctrl.h"
/*
class MemError
{
public:
  MemError() {};
};
*/
class Subset						// subsets are coded on bits
{
public:
  Subset() { a=NULL; };
  Subset(int);						// size in bits
  Subset(Subset&);
  ~Subset();

  void Zero();
  void Full();						// set all
  void Set(int k)					// set a bit
  { if((k<0)||(k>=n)) cerr << "Subset: set " << k << endl;
	 a[k/32]|=mask[k%32]; }

  void Set(int k,int val)		// sets a bit to value
  { if(val) Set(k);
	 else Reset(k); }

  void Reset(int k)				// resets a bit
  { if((k<0)||(k>=n)) cerr << "Subset: reset " << k << endl;
	 a[k/32]&=~(mask[k%32]); }

  int Ones();				  	   // no. of "1"
  int N() { return n; };		// no. of bits

  int operator[](int n)			// returns bit
  { return (a[n/32]&mask[n%32]?1:0); };
  int Val(int n)
  { return (a[n/32]&mask[n%32]?1:0); };
  int operator==(Subset&);
  int operator<(Subset&);		// lexicographic order
  int operator<=(Subset&);		// implication = inclusion
  Subset& operator=(Subset&);

  void Set(int*);	// a list of bits to set, finished by -1

  friend ostream& operator<<(ostream&,Subset&);

private:
  unsigned int *a;
  int nr;				// no. of bytes
  int n;					// no. of bits

  static unsigned int mask[32];
};

//**************************************************************

class SubsetTree
{
public:
  SubsetTree(Subset &);
  ~SubsetTree();

  int AddIfNotExists(Subset &);  // returns 1 if found, adds otherwise
  int Exists(Subset &);  			// returns 1 if found, does not add

  Subset x;
  SubsetTree *left,*right;
};

#endif
