#include "tabsort.h"

//************************************************

TDTableSort::TDTableSort(TDTable *tt)
{
	dec=NULL;
	t=tt;
	dec_dictionary=NULL;
	max_real_dec=0;
	
	bubble_sort_limit=12;
	last_sort_by=NULL;
	
	order=new int [NoObj()];
	for(int i=0;i<NoObj();i++) order[i]=i;
	o_n=new int * [NoAttr()];
	no_of_distinct=new int [NoAttr()];
	for(i=0;i<NoAttr();i++) 
	{
		o_n[i]=NULL;
		no_of_distinct[i]=0;
	}
	last_sort_by=new int [NoAttr()+1];
	for(i=0;i<NoAttr()+1;i++) last_sort_by[i]=-1;
	
	attributes=Subset(NoAttr());
};

TDTableSort::~TDTableSort()
{
	delete [] order;
	delete [] no_of_distinct;
	delete [] dec;
	delete [] last_sort_by;
	delete dec_dictionary;
	for(int i=0;i<NoAttr();i++) delete [] o_n[i];
	delete [] o_n;
}


void TDTableSort::SetDecision()
{
	Subset a(NoAttr());
	a.Full();
	a.Reset(NoAttr()-1);
	SetDecision(NoAttr()-1,a);
}

double TDTableSort::DecisionStatistics(int i)
{
	if(i==-1) return dec_dictionary->sum_stat;
	return dec_dictionary->statistics[i];
}

void TDTableSort::SetDecision(int d,Subset &a)
{
	t->SetDecAttr(d);
	delete [] dec;
	delete dec_dictionary;
	dec_dictionary=new GenDecision();
	dec_dictionary->SetDecNo(GetNoOfDistinct(d));
	dec=new int [NoObj()];
	attributes=a;
	int *ord_by=new int [NoAttr()+2];
	int i,k=0;
	for(i=0;i<NoAttr();i++) if(a[i]) ord_by[k++]=i;
	ord_by[k]=d;    // set sorting order to all attr. and decision on the end
	ord_by[k+1]=-1;
	Sort(ord_by);
	ord_by[k]=-1;	// this order will be used for comparing objects
	
	for(i=0;i<NoObj();i++)	// +1, because we have to check the last class too
	{
		dec_dictionary->AddValue(GetSortedTable(i,d));
		dec[Order(i)]=dec_dictionary->GetCodedDec(GetSortedTable(i,d));
	}
}

int TDTableSort::GetNoOfDistinct(int a)
{
	if(no_of_distinct[a]>0) return no_of_distinct[a];
	int ord[]={-1,-1};
	ord[0]=a;
	Sort(ord);
	no_of_distinct[a]=1;
	for(int o=1;o<NoObj();o++)
	{
		if(GetSortedTable(o,a)!=GetSortedTable(o-1,a)) 
			no_of_distinct[a]++;
	}
	return no_of_distinct[a];
}

int TDTableSort::GetIndex(int attr,int v)
{	// returns a maximal index of value not more than v, or -1.
	int ord[]={-1,-1};
	ord[0]=attr;
	Sort(ord);
	if(GetSortedTable(0,attr)>v) return -1;
	int i=NoObj()/2,step=i/2,prev=-1;
	int cur;
	do
	{
		cur=GetSortedTable(i,attr);
		if(cur<v) 
		{
			if(i==NoObj()-1 || GetSortedTable(i+1,attr)>v) break;
			i+=step;
		}
		else if(cur>v)
		{
			if(GetSortedTable(i-1,attr)<v) 
			{
				i=i-1;
				break;
			}
			i-=step;
		}
		if(i<0) i=0;
		if(i>=NoObj()) i=NoObj()-1;
		step=(step+1)/2;
	}
	while(cur!=v);
	while(i<NoObj()-1 && GetSortedTable(i+1,attr)<=v) i++;
	return i;
}

int TDTableSort::ExistsBetween(int attr,int v1,int v2)
{
	int i=GetIndex(attr,v1);
	if(i==NoObj()-1 || GetSortedTable(i+1,attr)>=v2) return 0;
	return 1;
}


void TDTableSort::Sort(int *sort_by)
{
	Sort(sort_by,0,NoObj()-1);
}

void TDTableSort::Sort(int *sort_by,int min,int max)
{
	int all=(min==0 && max==NoObj()-1);
	for(int i=0;sort_by[i]==last_sort_by[i] && sort_by[i]>-1; i++);
	// compare the current sequence and the beginning of the previous one
	if(all && sort_by[i]==-1) return;   // if equal, we don't need to sort
	
	int change=i;		// the first position changed
	int *compare=new int [NoAttr()+1];
	// this part was not changed
	for(i=0;i<change;i++) compare[i]=sort_by[i];
	compare[i]=-1;
	
	int first=sort_by[0];	//the first column to sort by
	if(all==0)
	{
		QuickSortTable(min,max,sort_by);
	}
	else if(o_n[first]==NULL)	// if there is no order for the first column,
	{
		QuickSortTable(0,NoObj()-1,sort_by);
		o_n[first]=new int [NoObj()];	// create it
		for(i=0;i<NoObj();i++)
		{
			o_n[first][i]=order[i];
		}
	}
	else if(sort_by[1]==-1)
	{
		for(i=0;i<NoObj();i++) order[i]=o_n[first][i];	
		// if we have to sort by just one column
	}	 // use stored order
	else if(change<2)
	{                // if there is no common beginning...
		for(i=0;i<NoObj();i++) order[i]=o_n[first][i];
		int block_start=0;     // sort by the first column
		for(int obj=min+1;obj<=max;obj++)
		{
			if(GetTable(order[obj-1],first)!=GetTable(order[obj],first))
			{		// then sort only these classes which have more
				if(obj-block_start>1)    // than one different value
					QuickSortTable(block_start,obj-1,sort_by+1);
				block_start=obj;		// sort by the second, third... column
			}
		}
		if(obj-block_start>1) QuickSortTable(block_start,max,sort_by+1);
	}
	else
	{	// if there was common beginning of column numbers
		int block_start=0;
		for(int obj=min+1;obj<=max;obj++)
		{
			if(Compare(order[obj-1],order[obj],compare)!=0)
			{           // sort only classes with more than one
				if(obj-block_start>1)		// value combination on the common columns
					QuickSortTable(block_start,obj-1,sort_by+change);
				block_start=obj;
			}
		}
		if(obj-block_start>1) QuickSortTable(block_start,max,sort_by+change);
	}
	// store the current order of attribs
	if(all)
	{
		for(i=0;sort_by[i]>-1;i++) last_sort_by[i]=sort_by[i];
		last_sort_by[i]=-1;
	}
	else last_sort_by[0]=-1;
	delete [] compare;
}

int TDTableSort::Compare(int o1,int o2,int *compare_with)
{
	for(int i=0;(i<NoAttr())&&(compare_with[i]!=-1);i++)
	{
		int w=GetTable(o1,compare_with[i])-GetTable(o2,compare_with[i]);
		if(w<0)	return 1;
		if(w>0)	return -1;
	}
	return 0;
}

int TDTableSort::Compare(int o1,int *o2,int *compare_with)
{
	for(int i=0;(i<NoAttr())&&(compare_with[i]!=-1);i++)
	{
		int w=GetTable(o1,compare_with[i])-o2[compare_with[i]];
		if(w<0)	return 1;
		if(w>0)	return -1;
	}
	return 0;
}

//***********************************************************

void TDTableSort::QuickSortTable(int l,int p,int *sort_by)
// Quick Sort - see Wirth "Algorithms+data struct=prog", page 90.
{
	int srodek=order[(l+p)/2];
	int w,i=l,j=p;
	do
	{
		while(Compare(order[i],srodek,sort_by)==1) i++;
		while(Compare(srodek,order[j],sort_by)==1) j--;
		if(i<=j)
		{
			w=order[i];
			order[i]=order[j];
			order[j]=w;
			i++;
			j--;
		};
	} while(i<=j);
	if(l<j)
	{
		if(j-l<bubble_sort_limit)
			BubbleSortTable(l,j,sort_by);
		else QuickSortTable(l,j,sort_by);
	}
	if(i<p)
	{
		if(p-i<bubble_sort_limit)
			BubbleSortTable(i,p,sort_by);
		else QuickSortTable(i,p,sort_by);
	}
}

void TDTableSort::BubbleSortTable(int l,int p,int *sort_by)
{
	int ost,pop_ost=p;
	do
	{
		ost=-1;
		for(int i=l;i<pop_ost;i++)
			if(Compare(order[i],order[i+1],sort_by)==-1)
			{
				int pom=order[i];
				order[i]=order[i+1];
				order[i+1]=pom;
				ost=i;
			}
			pop_ost=ost;
	}
	while(ost>-1);
}

//********************************************************