/*********************************************************

Name: tabsort.h
Author: JW
Revised by:
Version: 1.0.008
Date of mod.: 29.04.1998
Status: ready

Description:
Addition for TDTable object: a quick-sorter

Comments:
Limited version - no generalized decision (statistics only).
Modified RS2000 version.

  WARNING: probably unstable!
*********************************************************/

#ifndef _TABSORT_H_
#define _TABSORT_H_

#include "tdtable.h"
#include "subset.h"
#include "gen_dec.h"


class TDTableSort
{
public:
  TDTableSort(TDTable*);
  ~TDTableSort();

  void Sort(int*);	// input: an order of attributes to sort with:
							// e.g.: { 3, 0, 5, -1, -1, -1 }. Table must
							// have not more than NoAttr numbers (-1 - not sorted)
  void Sort(int *sort_by,int min,int max);	// sort only part of table
  int GetNoOfDistinct(int a);	// number of distinct values of attr.

  int GetSortedTable(int o,int a)
  { return t->GetTable(Order(o),a); }
  int GetTable(int o,int a)
  { return t->GetTable(o,a); }

  int *GetTable(int o)
  { return t->GetTable(o); }

  int GetIndex(int attr,int v);
	// returns a maximal index of value not more than v, or -1.
  int ExistsBetween(int attr,int v1,int v2);
					// return 1 iff exists a value between v1 and v2 (excl.)
  
  int NoObj()
  { return t->NoObj(); }
  int NoAttr()
  { return t->NoAttr(); }

  int Compare(int o1,int o2,int *compare_with);		// both from table
  int Compare(int o1,int *o2,int *compare_with);	// one external
							// Result: -1 - first is greater
							//				0 - equal
							//				1 - second is greater
							// compare_with - see Sort()
  int Order(int i)
  { return order[i]; };

  // Decision tools

  void SetDecision(int,Subset&);
			// parameters: number of decision attr., Subset of attr.
  void SetDecision();
			// default: decision is the last column
  double DecisionStatistics(int i);	// returns a statistics for i-th dec. class
  int GetDecAttr()
  { return t->GetDecAttr(); }
  int TranslateDec(int i)
  { return dec_dictionary->GetOrigDec(i); }
  int GetNoDecVals()
  { return dec_dictionary->GetDecNumber(); }
  int *GetDecVals()
  { return dec_dictionary->GetOrigDec(); }
  int *dec;
  Subset attributes;

protected:
  int *order;
  int **o_n;

private:
  int bubble_sort_limit;
  int *last_sort_by;
  int *no_of_distinct;
  TDTable *t;

  GenDecision *dec_dictionary;
  int max_real_dec;	// the last used number of non-gen. dec. value

  void QuickSortTable(int,int,int*);
  void BubbleSortTable(int,int,int*);
};


#endif